#######################################################################
# Copyright (c) 2014 ENEO Tecnologia S.L.
# This file is part of redBorder.
# redBorder is free software: you can redistribute it and/or modify
# it under the terms of the GNU Affero General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
# redBorder is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Affero General Public License for more details.
# You should have received a copy of the GNU Affero General Public License
# along with redBorder. If not, see <http://www.gnu.org/licenses/>.
#######################################################################
module Redborder
  module Jobs
    # This Jobs check alarms every 5 minutes and notify users when some of them
    # must be fired.
    class CheckAlarmsJob < Struct.new(:verbose)
      include Redborder::Jobs::JobHelper

      def perform
        alarms = Alarm.all.select(&:must_be_checked?)
      
        alarms.each do |alarm|
          # Check if the creator of the alarm can use this module
          next unless module_available? alarm

          data_source = load_data_source(alarm)
          filter = Redborder::Filter.new(data_source)
          filter.parse(alarm.condition) if alarm.condition
          
          start_time = Time.now - alarm.time_window.to_i
          end_time = Time.now
          # Check the alarm
          result = check alarm
          next unless result[:fired]

          # Check if is an alarm with incident_enabled, if incident dont exists we created
          if result[:fired]
            if alarm.incident_enabled && alarm.sensor
              if alarm.regeneration_mode.present?
                alarm.regenerate_incident_if_needed
              elsif alarm.incident.nil?
                incident = Incident.new
                incident.name = alarm.name
                priority = map_severity_to_vault_priority(alarm.severity.name)
                incident.priority = priority || Priority.find_by(name: 'unknown')
                incident.uuid = SecureRandom.uuid
                incident.incident_status = IncidentStatus.find_by(id: 4)  # 4 = NEW
                incident.source = Redborder::Module.registered_modules[:vault][:incident_source_name]
                incident.sensor_id = alarm.sensor.id
                if incident.save
                  alarm.incident = incident
                  alarm.save
                else
                  Rails.logger.error "Cannot create incident from check alarm job"
                end
              end
            end
      
            if alarm.incident_enabled
              more_info = AlarmsHelper::query_data(data_source, start_time, end_time, filter, alarm)
              processed_messages = AlarmsHelper.process_query_data({ more_info: more_info }, alarm)
              if processed_messages.is_a?(Array)
                processed_messages.each do |msg|
                  notifier = Redborder::AlarmNotificator.new(alarm, result.merge(more_info: msg))
                  notifier.notify
                end
              else
                result.merge!(more_info: processed_messages)
                Redborder::AlarmNotificator.new(alarm, result).notify
              end
            else
              Redborder::AlarmNotificator.new(alarm, result).notify
            end
            
            alarm.launch
          end
        end
      
        Delayed::Job.enqueue(
          Redborder::Jobs::CheckAlarmsJob.new(true), priority: 10, run_at: Time.now + 5.minutes)
      end   

      private

      def module_available?(alarm)
        return false unless alarm.sensor
        return true if alarm.sensor.modules.include? alarm.product_type
        logit "Module #{alarm.product_type} is not available for #{alarm.sensor.name}"\
              "(#{alarm.sensor.id})"
        false
      end

      # Return the datasource of the current module or nil if it isn't available
      #
      # @param alarm [Alarm] Alarm to check
      # @return [DataSource, nil] DataSource of the module
      #
      def load_data_source(alarm)
        # Return the datasource
        Redborder::DruidSource.get(source: alarm.product_type)
      rescue NotImplementedError
        # The mdoule hasn't got any datasource
        logit "The module #{alarm.product_type} hasn't got any DataSource"
        nil
      end

      # Check if we must fire the given alarm.
      #
      # @param alarm [Alarm] Alarm to check
      # @return [Hash]
      #
      def check(alarm)
        data_source = load_data_source(alarm)
        # Return if the datasource doesn't exist
        return { fired: false } unless data_source

        # Check for this alarm
        filter = Redborder::Filter.new(data_source)
        filter.parse(alarm.condition) if alarm.condition
        start_time = Time.now - alarm.time_window.to_i
        end_time = Time.now
        data = query(data_source, start_time, end_time, filter, alarm)
        # Check condition and return the result

        check_conditions(data, alarm).merge(start_time: start_time, end_time: end_time)
      end
      # Perform the query to data source to check the alarm
      #
      # @param data_source [DataSource] Class to perform the query
      # @param start_time [Time] Start time
      # @param end_time [Time] End time
      # @param filter [Filter] Filters to apply
      # @param alarm [Alarm] The alarm to fetch the user and aggregation
      # @return [Array] Result of the query
      #
      def query(data_source, start_time, end_time, filter, alarm)
          data_source.where(
            query_type: :timeseries,
            range: :custom,
            start_time: start_time,
            end_time: end_time,
            granularity: alarm.outliers_enabled ? alarm.outliers_granularity.downcase.to_sym : :all,
            search: filter.get,
            aggregation: alarm.aggregation,
            user: User.first,
            outliers: alarm.outliers_enabled
          )
      end

      # Check if the data meet the conditions
      #
      # @param data [Array] Result of the DataSource query
      # @param alarm [Alarm] Alarm to check the conditions
      # @return [Hash] Hash with result
      #
      def check_conditions(data, alarm)
        # Check conditions
          fired = false
          where = :nowhere
          if !alarm.outliers_enabled?
            value = data.empty? ? 0 : data.first.send(alarm.aggregation)
            if alarm.upper_limit_enabled? && value >= alarm.upper_mul_limit
              where = :upper
              fired = true
            elsif alarm.bottom_limit_enabled? && value <= alarm.bottom_mul_limit
              where = :bottom
              fired = true
            end
          else
            begin
              anomalies = JSON.parse(data.last.to_json)['anomalies']
              if !anomalies.empty?
                where = :sherlock
                fired = true
                value = anomalies.length
              end
            rescue
              #Sherlock service not enabled, incorrect HTTP response, etc...
              where = :nowhere
              fired = false
              value = 0
            end
          end
        # Return the result
        {
          where: where,
          value: value,
          fired: fired
        }
      end

      def map_severity_to_vault_priority(severity_name)
        case severity_name
        when 'very low'
          return Priority.find_by_name('debug')
        when 'low'
          return Priority.find_by_name('info')
        when 'medium'
          return Priority.find_by_name('notice')
        when 'high'
          return Priority.find_by_name('warning')
        when 'very high'
          return Priority.find_by_name('error')
        when 'critical'
          return Priority.find_by_name('critical')
        else
          return Priority.find_by_name('unknown')
        end
      end
    end
  end
end
